#include <iostream>
#include <vector>

#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>

#include <Eigen/Dense>

#include "StringEnumerator.hpp"
#include "PointCorrespondanceLoader.hpp"
#include "MultiRectification.hpp"

const bool verboseLM = true;



int main(int argc, char ** argv)
{
    // vector of images
    std::vector<cv::Mat> images;

    // load images
    std::cout << "load images ..." << std::endl;
    images.push_back(cv::imread("../input/image0.jpg"));
    images.push_back(cv::imread("../input/image1.jpg"));
    images.push_back(cv::imread("../input/image2.jpg"));
    images.push_back(cv::imread("../input/image3.jpg"));

    // get image size
    std::vector<std::pair<uint,uint>> imageSize;
    for(const auto & im : images)
        imageSize.push_back(std::pair<uint,uint>(im.cols, im.rows)); // (width, height)
    if(verboseLM) for(uint i=0; i < imageSize.size(); ++i) std::cout << "   image " << i << " : " << imageSize[i].first << " x " << imageSize[i].second << std::endl;

    // load (or compute by yourself) point correspondances
    std::cout << "load point correspondances ..." << std::endl;
    std::vector<Eigen::MatrixXd> correspondences;
    correspondences.push_back(load_point_corrspondances("../input/image0.list"));
    correspondences.push_back(load_point_corrspondances("../input/image1.list"));
    correspondences.push_back(load_point_corrspondances("../input/image2.list"));
    correspondences.push_back(load_point_corrspondances("../input/image3.list"));
    if(verboseLM) for(const auto &vec_list : correspondences) std::cout << "----- nb points: " << vec_list.rows() << std::endl << vec_list << std::endl << std::endl;

    // display input images
    for(unsigned int im=0; im<images.size(); ++im){
        // draw points
        for(unsigned int i=0; i<correspondences[im].rows(); ++i)
            cv::circle(images[im], cv::Point2i(correspondences[im](i,0),correspondences[im](i,1)), 3, cv::Scalar(0,0,255), -1);
        // display image
        cv::imshow(stringEnumerator("image", im, 2), images[im]);
        //cv::imwrite(stringEnumerator("/tmp/image_", im, 2,".jpg"),images[im]);
    }
    cv::waitKey(100);

    // image rectification
    std::cout << correspondences[0].rows() << " correspondences" << std::endl;
    std::cout << "rectification computation ..." << std::endl;
    const uint referenceImage = images.size()/2; // center image is better, but any other image is fine
    MultiRectification multi(correspondences,imageSize,referenceImage);
    std::vector<Eigen::Matrix3d> H;            // rectifying homographies
    std::vector<Eigen::Matrix<double,3,4>> P;  // projection matrices
    multi.minimize(H,P,300000,verboseLM);

    // show result
    if(verboseLM) for(const auto &h: H) std::cout << "Homography:\n" << h << std::endl << std::endl;
    if(verboseLM) for(const auto &p: P) std::cout << "P :\n" << p << std::endl << std::endl;

    // apply homographies
    for(uint im=0; im<images.size(); ++im){
        // convert homography from Eigen to OpenCV
        cv::Mat Hocv(3,3,CV_64F);
        for(uint i=0; i<3; ++i)
            for(uint j=0; j<3; ++j)
                Hocv.at<double>(i,j) = H[im](i,j);

#if 1 // optional (center final image in a bigger frame)
        const int scaleFactor = 2;
        cv::Mat centerTranslation = (cv::Mat_<double>(3,3) << 1, 0, 0.5*scaleFactor*images[im].cols/2, 0, 1, 0.5*scaleFactor*images[im].rows/2, 0 , 0, 1);
        Hocv = centerTranslation * Hocv;
        cv::warpPerspective(images[im], images[im], Hocv, images[im].size()*scaleFactor);
#else
        cv::warpPerspective(images[im], images[im], Hocv, images[im].size() );
#endif
        cv::imshow(stringEnumerator("image", im, 2), images[im]);
        cv::imwrite(stringEnumerator("/tmp/imageRes_", im, 2,".jpg"),images[im]);
    }
    cv::waitKey(1000);


    return 0;
}